#!/usr/bin/env python

from __future__ import print_function
import re
import sys

re_identifier = re.compile(r'^[a-zA-Z_]')
re_compat_handle = re.compile(r'^COMPAT_HANDLE\((.*)\)$')
re_pad = re.compile(r'^_pad\d*$')
re_compat = re.compile(r'^compat_.*_t$')

def removeprefix(s, prefix):
    if s.startswith(prefix):
        return s[len(prefix):]
    return s

def removesuffix(s, suffix):
    if s.endswith(suffix):
        return s[:-len(suffix)]
    return s

def get_fields(looking_for, header_tokens):
    level = 1
    aggr = 0
    fields = []
    name = ''

    for token in header_tokens:
        if token in ('struct', 'union'):
            if level == 1:
                aggr = 1
                fields = []
                name = ''
        elif token == '{':
            level += 1
        elif token == '}':
            level -= 1
            if level == 1 and name == looking_for:
                fields.append(token)
                return fields
        elif re_identifier.match(token):
            if not (aggr == 0 or name != ''):
                name = token

        if aggr != 0:
            fields.append(token)

    return []

def build_enums(name, tokens):
    level = 1
    kind = ''
    named = ''
    fields = []
    members = []
    id = ''

    for token in tokens:
        if token in ('struct', 'union'):
            if not level != 2:
                fields = ['']
            kind = "%s;%s" % (token, kind)
        elif token == '{':
            level += 1
        elif token == '}':
            level -= 1
            if level == 1:
                subkind = kind
                (subkind, _, _) = subkind.partition(';')
                if subkind == 'union':
                    print("\nenum XLAT_%s {" % (name,))
                    for m in members:
                        print("    XLAT_%s_%s," % (name, m))
                    print("};")
                return
            elif level == 2:
                named = '?'
        elif re_identifier.match(token):
            id = token
            k = kind
            (_, _, k) = k.partition(';')
            if named != '' and k != '':
                if len(fields) > 0 and fields[0] == '':
                    fields.pop(0)
                build_enums("%s_%s" % (name, token), fields)
                named = '!'
        elif token == ',':
            if level == 2:
                members.append(id)
        elif token == ';':
            if level == 2:
                members.append(id)
            if named != '':
                (_, _, kind) = kind.partition(';')
            named = ''
        if len(fields) != 0:
            fields.append(token)

def handle_field(prefix, name, id, type, fields):
    if len(fields) == 0:
        print(" \\")
        if type == '':
            print("%s(_d_)->%s = (_s_)->%s;" % (prefix, id, id), end='')
        else:
            k = id.replace('.', '_')
            print("%sXLAT_%s_HNDL_%s(_d_, _s_);" % (prefix, name, k), end='')
    elif not '{' in fields:
        tag = ' '.join(fields)
        tag = re.sub(r'\s*(struct|union)\s+(compat_)?(\w+)\s.*', '\\3', tag)
        print(" \\")
        print("%sXLAT_%s(&(_d_)->%s, &(_s_)->%s);" % (prefix, tag, id, id), end='')
    else:
        func_id = id
        func_tokens = fields
        kind = ''
        array = ""
        level = 1
        arrlvl = 1
        array_type = ''
        id = ''
        type = ''
        fields = []
        for token in func_tokens:
            if token in ('struct', 'union'):
                if level == 2:
                    fields = ['']
                if level == 1:
                    kind = token
                    if kind == 'union':
                        tmp = func_id.replace('.', '_')
                        print(" \\")
                        print("%sswitch (%s) {" % (prefix, tmp), end='')
            elif token == '{':
                level += 1
                id = ''
            elif token == '}':
                level -= 1
                id = ''
                if level == 1 and kind == 'union':
                    print(" \\")
                    print("%s}" % (prefix,), end='')
            elif token == '[':
                if level != 2 or arrlvl != 1:
                    pass
                elif array == '':
                    array = ' '
                else:
                    array = "%s;" % (array,)
                arrlvl += 1
            elif token == ']':
                arrlvl -= 1
            elif re_compat_handle.match(token):
                if level == 2 and id == '':
                    m = re_compat_handle.match(token)
                    type = m.groups()[0]
                    type = removeprefix(type, 'compat_')
            elif token == "compat_domain_handle_t":
                if level == 2 and id == '':
                    array_type = token
            elif re_identifier.match(token):
                id = token
            elif token in (',', ';'):
                if level == 2 and not re_pad.match(id):
                    if kind == 'union':
                        tmp = "%s.%s" % (func_id, id)
                        tmp = tmp.replace('.', '_')
                        print(" \\")
                        print("%scase XLAT_%s_%s:" % (prefix, name, tmp), end='')
                        if len(fields) > 0 and fields[0] == '':
                            fields.pop(0)
                        handle_field("%s    " % (prefix,), name, "%s.%s" % (func_id, id), type, fields)
                    elif array == '' and array_type == '':
                        if len(fields) > 0 and fields[0] == '':
                            fields.pop(0)
                        handle_field(prefix, name, "%s.%s" % (func_id, id), type, fields)
                    elif array == '':
                        copy_array("    ", "%s.%s" % (func_id, id))
                    else:
                        (_, _, array) = array.partition(';')
                        if len(fields) > 0 and fields[0] == '':
                            fields.pop(0)
                        handle_array(prefix, name, "{func_id}.{id}", array, type, fields)
                    if token == ';':
                        fields = []
                        id = ''
                        type = ''
                    array = ''
                    if kind == 'union':
                        print(" \\")
                        print("%s    break;" % (prefix,), end='')
            else:
                if array != '':
                    array = "%s %s" % (array, token)
            if len(fields) > 0:
                fields.append(token)

def copy_array(prefix, id):
    print(" \\")
    print("%sif ((_d_)->%s != (_s_)->%s) \\" % (prefix, id, id))
    print("%s    memcpy((_d_)->%s, (_s_)->%s, sizeof((_d_)->%s));" % (prefix, id, id, id), end='')

def handle_array(prefix, name, id, array, type, fields):
    i = re.sub(r'[^;]', '', array)
    i = "i%s" % (len(i),)

    print(" \\")
    print("%s{ \\" % (prefix,))
    print("%s    unsigned int %s; \\" % (prefix, i))
    (head, _, tail) = array.partition(';')
    head = head.strip()
    print("%s    for (%s = 0; %s < %s; ++%s) {" % (prefix, i, i, head, i), end='')
    if not ';' in array:
        handle_field("%s        " % (prefix,), name, "%s[%s]" % (id, i), type, fields)
    else:
        handle_array("%s        " % (prefix,) , name, "%s[%s]" % (id, i), tail, type, fields)
    print(" \\")
    print("%s    } \\" % (prefix,))
    print("%s}" % (prefix,), end='')

def build_body(name, tokens):
    level = 1
    id = ''
    array = ''
    arrlvl = 1
    array_type = ''
    type = ''
    fields = []

    print("\n#define XLAT_%s(_d_, _s_) do {" % (name,), end='')

    for token in tokens:
        if token in ('struct', 'union'):
            if level == 2:
                fields = ['']
        elif token == '{':
            level += 1
            id = ''
        elif token == '}':
            level -= 1
            id = ''
        elif token == '[':
            if level != 2 or arrlvl != 1:
                pass
            elif array == '':
                array = ' '
            else:
                array = "%s;" % (array,)
            arrlvl += 1
        elif token == ']':
            arrlvl -= 1
        elif re_compat_handle.match(token):
            if level == 2 and id == '':
                m = re_compat_handle.match(token)
                type = m.groups()[0]
                type = removeprefix(type, 'compat_')
        elif token == "compat_domain_handle_t":
            if level == 2 and id == '':
                array_type = token
        elif re_identifier.match(token):
            if array != '':
                array = "%s %s" % (array, token)
            else:
                id = token
        elif token in (',', ';'):
            if level == 2 and not re_pad.match(id):
                if array == '' and array_type == '':
                    if len(fields) > 0 and fields[0] == '':
                        fields.pop(0)
                    handle_field("    ", name, id, type, fields)
                elif array == '':
                    copy_array("    ", id)
                else:
                    (head, sep, tmp) = array.partition(';')
                    if sep == '':
                        tmp = head
                    if len(fields) > 0 and fields[0] == '':
                        fields.pop(0)
                    handle_array("    ", name, id, tmp, type, fields)
                if token == ';':
                    fields = []
                    id = ''
                    type = ''
                array = ''
        else:
            if array != '':
                array = "%s %s" % (array, token)
        if len(fields) > 0:
            fields.append(token)
    print(" \\\n} while (0)")

def check_field(kind, name, field, extrafields):
    if not '{' in extrafields:
        print("; \\")
        if len(extrafields) != 0:
            for token in extrafields:
                if token in ('struct', 'union'):
                    pass
                elif re_identifier.match(token):
                    print("    CHECK_%s" % (removeprefix(token, 'xen_'),), end='')
                    break
                else:
                    raise Exception("Malformed compound declaration: '%s'" % (token,))
        elif not '.' in field:
            print("    CHECK_FIELD_(%s, %s, %s)" % (kind, name, field), end='')
        else:
            n = field.count('.')
            field = field.replace('.', ', ')
            print("    CHECK_SUBFIELD_%s_(%s, %s, %s)" % (n, kind, name, field), end='')
    else:
        level = 1
        fields = []
        id = ''

        for token in extrafields:
            if token in ('struct', 'union'):
                if level == 2:
                    fields = ['']
            elif token == '{':
                level += 1
                id = ''
            elif token == '}':
                level -= 1
                id = ''
            elif re_compat.match(token):
                if level == 2:
                    fields = ['']
                    token = removesuffix(token, '_t')
                    token = removeprefix(token, 'compat_')
            elif re.match(r'^evtchn_.*_compat_t$', token):
                if level == 2 and token != "evtchn_port_compat_t":
                    fields = ['']
                    token = removesuffix(token, '_compat_t')
            elif re_identifier.match(token):
                id = token
            elif token in (',', ';'):
                if level == 2 and not re_pad.match(id):
                    if len(fields) > 0 and fields[0] == '':
                        fields.pop(0)
                    check_field(kind, name, "%s.%s" % (field, id), fields)
                    if token == ";":
                        fields = []
                        id = ''
            if len(fields) > 0:
                fields.append(token)

def build_check(name, tokens):
    level = 1
    fields = []
    kind = ''
    id = ''
    arrlvl = 1

    print("")
    print("#define CHECK_%s \\" % (name,))

    for token in tokens:
        if token in ('struct', 'union'):
            if level == 1:
                kind = token
                print("    CHECK_SIZE_(%s, %s)" % (kind, name), end='')
            elif level == 2:
                fields = ['']
        elif token == '{':
            level += 1
            id = ''
        elif token == '}':
            level -= 1
            id = ''
        elif token == '[':
            arrlvl += 1
        elif token == ']':
            arrlvl -= 1
        elif re_compat.match(token):
            if level == 2 and token != "compat_argo_port_t":
                fields = ['']
                token = removesuffix(token, '_t')
                token = removeprefix(token, 'compat_')
        elif re_identifier.match(token):
            if not (level != 2 or arrlvl != 1):
                id = token
        elif token in (',', ';'):
            if level == 2 and not re_pad.match(id):
                if len(fields) > 0 and fields[0] == '':
                    fields.pop(0)
                check_field(kind, name, id, fields)
                if token == ";":
                    fields = []
                    id = ''

        if len(fields) > 0:
            fields.append(token)
    print("")


def main():
    header_tokens = []
    re_tokenazier = re.compile(r'\s+')
    re_skip_line = re.compile(r'^\s*(#|$)')
    re_spacer = re.compile(r'([\]\[,;:{}])')

    with open(sys.argv[1]) as header:
        for line in header:
            if re_skip_line.match(line):
                continue
            line = re_spacer.sub(' \\1 ', line)
            line = line.strip()
            header_tokens += re_tokenazier.split(line)

    with open(sys.argv[2]) as compat_list:
        for line in compat_list:
            words = re_tokenazier.split(line, maxsplit=1)
            what = words[0]
            name = words[1]

            name = removeprefix(name, 'xen')
            name = name.strip()

            fields = get_fields("compat_%s" % (name,), header_tokens)
            if len(fields) == 0:
                raise Exception("Fields of 'compat_%s' not found in '%s'" % (name, sys.argv[1]))

            if what == "!":
                build_enums(name, fields)
                build_body(name, fields)
            elif what == "?":
                build_check(name, fields)
            else:
                raise Exception("Invalid translation indicator: '%s'" % (what,))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import os
from . import settings, utils, tag_database, cppcheck_analysis

class ParseTagPhaseError(Exception):
    pass

class BuildPhaseError(Exception):
    pass

class CleanPhaseError(Exception):
    pass


def parse_xen_tags():
    # Load the database for the Xen tags
    subs_list = tag_database.load_tag_database(
        settings.analysis_tool,
        [settings.repo_dir + "/docs/misra/safe.json"]
    )
    subs_list = tag_database.load_tag_database(
        settings.analysis_tool,
        [settings.repo_dir + "/docs/misra/false-positive-{}.json"
                                .format(settings.analysis_tool)],
        subs_list,
        "false-positive"
    )

    # Create outdir if it doesn't exists
    os.makedirs(settings.outdir, exist_ok=True)

    # The following lambda function will return a file if it contains lines with
    # a comment containing "SAF-<number>-{safe|false-positive-<tool>}" on a
    # single line.
    grep_action = lambda x: utils.grep(x,
                                    tag_database.get_xen_tag_comment_regex(
                                                        settings.analysis_tool)
    )
    # Look for a list of .h/.c files that matches the condition above
    parse_file_list = utils.recursive_find_file(settings.xen_dir, r'.*\.[ch]$',
                                                grep_action)

    for entry in parse_file_list:
        file = entry["file"]
        bkp_file = file + ".safparse"
        if os.path.isfile(bkp_file):
            raise ParseTagPhaseError(
                "Found {}, please check the integrity of {}"
                    .format(bkp_file,file)
                )
        os.rename(file, bkp_file)
        time_bkp_file = os.stat(bkp_file)
        # Create <file> from <file>.safparse but with the Xen tag parsed
        try:
            tag_database.substitute_tags(settings.analysis_tool, bkp_file, entry,
                                         subs_list)
        except Exception as e:
            raise ParseTagPhaseError("{}".format(e))
        finally:
            # Set timestamp for file equal to bkp_file, so that if the file is
            # modified during the process by the user, we can catch it
            os.utime(file, (time_bkp_file.st_atime, time_bkp_file.st_mtime))


def build_xen():
    utils.invoke_command(
            "make -C {} {} {} build"
                .format(settings.xen_dir, settings.make_forward_args,
                        cppcheck_analysis.cppcheck_extra_make_args),
            False, BuildPhaseError,
            "Build error occured when running:\n{}"
        )


def clean_analysis_artifacts():
    safparse_files = utils.recursive_find_file(settings.xen_dir,
                                               r'.*.safparse$')
    for original_file in safparse_files:
        # This commands strips the .safparse extension, leaving <file>
        parsed_file_path = os.path.splitext(original_file)[0]
        mtime_original_file = os.stat(original_file).st_mtime
        mtime_parsed_file = os.stat(parsed_file_path).st_mtime
        if mtime_original_file != mtime_parsed_file:
            return CleanPhaseError(
                    "The file {} was modified during the analysis "
                    "procedure, it is impossible now to restore from the "
                    "content of {}, please handle it manually"
                    .format(parsed_file_path, original_file)
                )
        # Replace <file>.safparse to <file>
        os.replace(original_file, parsed_file_path)



#!/usr/bin/env python3

import os, re, subprocess


def grep(filepath, regex):
    regObj = re.compile(regex)
    res = { "file": filepath, "matches": {} }
    try:
        with open(filepath, "rt") as f:
            line_number = 1
            for line in f:
                match = regObj.match(line)
                if match:
                    res["matches"][line_number] = match
                line_number = line_number + 1
    except Exception as e:
        print("WARNING: Can't open {}: {}".format(filepath, e))

    # Return filename and line matches if there are
    return res if res["matches"] else {}


def recursive_find_file(path, filename_regex, action = None):
    filename_reg_obj = re.compile(filename_regex)
    res = []
    for root, dirs, fnames in os.walk(path):
        for fname in fnames:
            if filename_reg_obj.match(fname):
                if action is None:
                    res.append(os.path.join(root, fname))
                else:
                    out = action(os.path.join(root, fname))
                    if out:
                        res.append(out)

    return res


def invoke_command(command, needs_output, exeption_type = Exception,
                   exeption_msg = ""):
    try:
        pipe_stdout = subprocess.PIPE if (needs_output == True) else None
        output = subprocess.run(command, shell=True, check=True,
                                stdout=pipe_stdout, stderr=subprocess.STDOUT,
                                encoding='utf8')
    except (subprocess.CalledProcessError, subprocess.SubprocessError) as e:
        if needs_output == True:
            exeption_msg = exeption_msg.format(e.cmd, output.stdout)
        else:
            exeption_msg = exeption_msg.format(e.cmd)
        excp = exeption_type(exeption_msg)
        excp.errorcode = e.returncode if hasattr(e, 'returncode') else 1
        raise excp

    return output.stdout



#!/usr/bin/env python3

import re
from .report import Report, ReportError


class CppcheckReport(Report):
    def __init__(self, report_path):
        # type: (str) -> None
        super(CppcheckReport, self).__init__(report_path)
        # This matches a string like:
        # path/to/file.c(<line number>,<digits>):<whatever>
        # and captures file name path and line number
        # the last capture group is used for text substitution in __str__
        self.__report_entry_regex = re.compile(r'^(.*)\((\d+)(,\d+\):.*)$')

    def parse(self):
        # type: () -> None
        report_path = self.get_report_path()
        try:
            with open(report_path, "rt") as infile:
                report_lines = infile.readlines()
        except OSError as e:
            raise ReportError("Issue with reading file {}: {}"
                              .format(report_path, e))
        for line in report_lines:
            entry = self.__report_entry_regex.match(line)
            if entry and entry.group(1) and entry.group(2):
                file_path = entry.group(1)
                line_number = int(entry.group(2))
                self.add_entry(file_path, line_number, line)
            else:
                raise ReportError("Malformed report entry in file {}:\n{}"
                                  .format(report_path, line))

    def __str__(self):
        # type: () -> str
        ret = ""
        for entry in self.to_list():
            ret += re.sub(self.__report_entry_regex,
                          r'{}({}\3'.format(entry.file_path,
                                            entry.line_number),
                          entry.text)
        return ret



#!/usr/bin/python

from __future__ import print_function

import sys, re
from structs import unions, structs, defines

# command line arguments
arch    = sys.argv[1]
outfile = sys.argv[2]
infiles = sys.argv[3:]


###########################################################################
# configuration #2: architecture information

inttypes = {}
header = {}
footer = {}

#arm
inttypes["arm32"] = [
    ("unsigned long", "__danger_unsigned_long_on_arm32"),
    ("long",          "__danger_long_on_arm32"),
    ("xen_pfn_t",     "uint64_t"),
    ("xen_ulong_t",   "uint64_t"),
    ("uint64_t",      "__align8__ uint64_t"),
]
header["arm32"] = """
#define __arm___ARM32 1
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
# define __DECL_REG(n64, n32) union { uint64_t n64; uint32_t n32; }
# define __align8__ __attribute__((aligned (8)))
#else
# define __DECL_REG(n64, n32) uint64_t n64
# define __align8__ FIXME
#endif
"""
footer["arm32"] = """
#undef __DECL_REG
"""

inttypes["arm64"] = [
    ("unsigned long", "__danger_unsigned_long_on_arm64"),
    ("long",          "__danger_long_on_arm64"),
    ("xen_pfn_t",     "uint64_t"),
    ("xen_ulong_t",   "uint64_t"),
    ("uint64_t",      "__align8__ uint64_t"),
]
header["arm64"] = """
#define __aarch64___ARM64 1
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
# define __DECL_REG(n64, n32) union { uint64_t n64; uint32_t n32; }
# define __align8__ __attribute__((aligned (8)))
#else
# define __DECL_REG(n64, n32) uint64_t n64
# define __align8__ FIXME
#endif
"""
footer["arm64"] = """
#undef __DECL_REG
"""

# x86_32
inttypes["x86_32"] = [
    ("unsigned long", "uint32_t"),
    ("long",          "uint32_t"),
    ("xen_pfn_t",     "uint32_t"),
    ("xen_ulong_t",   "uint32_t"),
]
header["x86_32"] = """
#define __DECL_REG_LO8(which) uint32_t e ## which ## x
#define __DECL_REG_LO16(name) uint32_t e ## name
#define __i386___X86_32 1
#pragma pack(4)
"""
footer["x86_32"] = """
#undef __DECL_REG_LO8
#undef __DECL_REG_LO16
#pragma pack()
"""

# x86_64
inttypes["x86_64"] = [
    ("unsigned long", "__align8__ uint64_t"),
    ("long",          "__align8__ uint64_t"),
    ("xen_pfn_t",     "__align8__ uint64_t"),
    ("xen_ulong_t",   "__align8__ uint64_t"),
]
header["x86_64"] = """
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
# define __DECL_REG(name) union { uint64_t r ## name, e ## name; }
# define __align8__ __attribute__((aligned (8)))
#else
# define __DECL_REG(name) uint64_t r ## name
# define __align8__ FIXME
#endif
#define __DECL_REG_LOHI(name) __DECL_REG(name ## x)
#define __DECL_REG_LO8        __DECL_REG
#define __DECL_REG_LO16       __DECL_REG
#define __DECL_REG_HI         __DECL_REG
#define __x86_64___X86_64 1
"""
footer["x86_64"] = """
#undef __DECL_REG
#undef __DECL_REG_LOHI
#undef __DECL_REG_LO8
#undef __DECL_REG_LO16
#undef __DECL_REG_HI
"""

###########################################################################
# main

input  = ""
output = ""
fileid = re.sub("[-.]", "_", "__FOREIGN_%s__" % outfile.upper())

for name in infiles:
    f = open(name, "r")

    # Sanity check the licence of the input file(s)
    line = f.readline()
    if line != "/* SPDX-License-Identifier: MIT */\n":
        print("Error: %s %s Missing or unexpected SPDX tag '%s'" %
              (sys.argv[0], name, line.strip()), file=sys.stderr)
        exit(1)

    input += f.read()
    f.close()

# replace path in "infiles" by path in '/usr/include' to avoid exposing the
# build directory path in the generated headers.
headers_name_list = ""
public_headers_location = 'xen/include/public/'
for name in infiles:
    i = name.rindex(public_headers_location)
    i += len(public_headers_location)
    headers_name_list += " xen/%s" % (name[i:])

# add header
output += """/* SPDX-License-Identifier: MIT */
/*
 * public xen defines and struct for %s
 * generated from%s by %s -- DO NOT EDIT
 */

#ifndef %s
#define %s 1

""" % (arch, headers_name_list, sys.argv[0], fileid, fileid)

if arch in header:
    output += header[arch]
    output += "\n"

defined = {}

# add defines to output
for line in re.findall("#define[^\n]+", input):
    for define in defines:
        regex = r"#define\s+%s\b" % define
        match = re.search(regex, line)
        if None == match:
            continue
        defined[define] = 1
        if define.upper()[0] == define[0]:
            replace = define + "_" + arch.upper()
        else:
            replace = define + "_" + arch
        regex = r"\b%s\b" % define
        output += re.sub(regex, replace, line) + "\n"
output += "\n"

# delete defines, comments, empty lines
input = re.sub("#define[^\n]+\n", "", input)
input = re.compile(r"/\*(.*?)\*/", re.S).sub("", input)
input = re.compile(r"\n\s*\n", re.S).sub("\n", input)

# add unions to output
for union in unions:
    regex = r"union\s+%s\s*\{(.*?)\n\};" % union
    match = re.search(regex, input, re.S)
    if None == match:
        output += "#define %s_has_no_%s 1\n" % (arch, union)
    else:
        output += "union %s_%s {%s\n};\n" % (union, arch, match.group(1))
    output += "\n"

# add structs to output
for struct in structs:
    regex = r"(?:#ifdef ([A-Z_]+))?\nstruct\s+%s\s*\{(.*?)\n\};" % struct
    match = re.search(regex, input, re.S)
    if None == match or \
           (match.group(1) is not None and match.group(1) not in defined):
        output += "#define %s_has_no_%s 1\n" % (arch, struct)
    else:
        output += "struct %s_%s {%s\n};\n" % (struct, arch, match.group(2))
        output += "typedef struct %s_%s %s_%s_t;\n" % (struct, arch, struct, arch)
    output += "\n"

# add footer
if arch in footer:
    output += footer[arch]
    output += "\n"
output += "#endif /* %s */\n" % fileid

# replace: defines
for define in defines:
    if define.upper()[0] == define[0]:
        replace = define + "_" + arch.upper()
    else:
        replace = define + "_" + arch
    output = re.sub(r"\b%s\b" % define, replace, output)

# replace: unions
for union in unions:
    output = re.sub(r"\b(union\s+%s)\b" % union, r"\1_%s" % arch, output)

# replace: structs + struct typedefs
for struct in structs:
    output = re.sub(r"\b(struct\s+%s)\b" % struct, r"\1_%s" % arch, output)
    output = re.sub(r"\b(%s)_t\b" % struct, r"\1_%s_t" % arch, output)

# replace: integer types
for old, new in inttypes[arch]:
    output = re.sub(r"\b%s\b" % old, new, output)

# print results
with open(outfile, "w") as f:
    f.write(output)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys, os, re

if sys.version_info < (3, 0):
    range = xrange

class Fail(Exception):
    pass

class State(object):

    def __init__(self, input, output):

        self.source = input
        self.input  = open_file_or_fd(input, "r", 2)
        self.output = open_file_or_fd(output, "w", 2)

        # State parsed from input
        self.names = {}  # Value => Name mapping
        self.values = {} # Name => Value mapping
        self.raw = {
            '!': set(), '|': set(),
            'A': set(), 'S': set(), 'H': set(),
            'a': set(), 's': set(), 'h': set(),
        }

        # State calculated
        self.nr_entries = 0 # Number of words in a featureset
        self.common_1d = 0 # Common features between 1d and e1d
        self.pv_def = set() # PV default features
        self.hvm_shadow_def = set() # HVM shadow default features
        self.hvm_hap_def = set() # HVM HAP default features
        self.pv_max = set() # PV max features
        self.hvm_shadow_max = set() # HVM shadow max features
        self.hvm_hap_max = set() # HVM HAP max features
        self.bitfields = [] # Text to declare named bitfields in C
        self.deep_deps = {} # { feature num => dependant features }
        self.nr_deep_deps = 0 # Number of entries in deep_deps
        self.deep_features = set() # featureset of keys in deep_deps

def parse_definitions(state):
    """
    Parse featureset information from @param f and mutate the global
    namespace with symbols
    """
    feat_regex = re.compile(
        r"^XEN_CPUFEATURE\(([A-Z0-9_]+),"
        r"\s+([\s\d]+\*[\s\d]+\+[\s\d]+)\)"
        r"\s+/\*([\w!|]*) .*$")

    word_regex = re.compile(
        r"^/\* .* word (\d*) \*/$")
    last_word = -1

    this = sys.modules[__name__]

    for l in state.input.readlines():

        # Short circuit the regexes...
        if not (l.startswith("XEN_CPUFEATURE(") or
                l.startswith("/* ")):
            continue

        # Handle /* ... word $N */ lines
        if l.startswith("/* "):

            res = word_regex.match(l)
            if res is None:
                continue # Some other comment

            word = int(res.groups()[0])

            if word != last_word + 1:
                raise Fail("Featureset word %u out of order (last word %u)"
                           % (word, last_word))

            last_word = word
            state.nr_entries = word + 1
            continue

        # Handle XEN_CPUFEATURE( lines
        res = feat_regex.match(l)

        if res is None:
            raise Fail("Failed to interpret '%s'" % (l.strip(), ))

        name = res.groups()[0]
        val = eval(res.groups()[1]) # Regex confines this to a very simple expression
        attr = res.groups()[2]

        if hasattr(this, name):
            raise Fail("Duplicate symbol %s" % (name,))

        if val in state.names:
            raise Fail("Aliased value between %s and %s" %
                       (name, state.names[val]))

        # Mutate the current namespace to insert a feature literal with its
        # bit index.  Prepend an underscore if the name starts with a digit.
        if name[0] in "0123456789":
            this_name = "_" + name
        else:
            this_name = name
        setattr(this, this_name, val)

        # Construct forward and reverse mappings between name and value
        state.names[val] = name
        state.values[name.lower().replace("_", "-")] = val

        for a in attr:
            try:
                state.raw[a].add(val)
            except KeyError:
                raise Fail("Unrecognised attribute '%s' for %s" % (a, name))

    if len(state.names) == 0:
        raise Fail("No features found")

    if state.nr_entries == 0:
        raise Fail("No featureset word info found")

    max_val = max(state.names.keys())
    if (max_val >> 5) >= state.nr_entries:
        max_name = state.names[max_val]
        raise Fail("Feature %s (%d*32+%d) exceeds FEATURESET_NR_ENTRIES (%d)"
                   % (max_name, max_val >> 5, max_val & 31, state.nr_entries))

def featureset_to_uint32s(fs, nr):
    """ Represent a featureset as a list of C-compatible uint32_t's """

    bitmap = 0
    for f in fs:
        bitmap |= 1 << f

    words = []
    while bitmap:
        words.append(bitmap & ((1 << 32) - 1))
        bitmap >>= 32

    assert len(words) <= nr

    if len(words) < nr:
        words.extend([0] * (nr - len(words)))

    return ("0x%08xU" % x for x in words)

def format_uint32s(state, featureset, indent):
    """ Format a list of uint32_t's suitable for a macro definition """
    words = featureset_to_uint32s(featureset, state.nr_entries)
    spaces = " " * indent
    return spaces + (", \\\n" + spaces).join(words) + ", \\"


def crunch_numbers(state):

    # Features common between 1d and e1d.
    common_1d = (FPU, VME, DE, PSE, TSC, MSR, PAE, MCE, CX8, APIC,
                 MTRR, PGE, MCA, CMOV, PAT, PSE36, MMX, FXSR)
    state.common_1d = common_1d

    state.pv_def =                                state.raw['A']
    state.hvm_shadow_def = state.pv_def         | state.raw['S']
    state.hvm_hap_def =    state.hvm_shadow_def | state.raw['H']

    state.pv_max =                                state.raw['A'] | state.raw['a']
    state.hvm_shadow_max = state.pv_max         | state.raw['S'] | state.raw['s']
    state.hvm_hap_max =    state.hvm_shadow_max | state.raw['H'] | state.raw['h']

    #
    # Feature dependency information.
    #
    # !!! WARNING !!!
    #
    # A lot of this information is derived from the written text of vendors
    # software manuals, rather than directly from a statement.  As such, it
    # does contain guesswork and assumptions, and may not accurately match
    # hardware implementations.
    #
    # It is however designed to create an end result for a guest which does
    # plausibly match real hardware.
    #
    # !!! WARNING !!!
    #
    # The format of this dictionary is that the feature in the key is a direct
    # prerequisite of each feature in the value.
    #
    # The first consideration is about which functionality is physically built
    # on top of other features.  The second consideration, which is more
    # subjective, is whether real hardware would ever be found supporting
    # feature X but not Y.
    #
    deps = {
        # FPU is taken to mean support for the x87 regisers as well as the
        # instructions.  MMX is documented to alias the %MM registers over the
        # x87 %ST registers in hardware.  Correct restoring of error pointers
        # of course makes no sense without there being anything to restore.
        FPU: [MMX, RSTR_FP_ERR_PTRS],

        # The PSE36 feature indicates that reserved bits in a PSE superpage
        # may be used as extra physical address bits.
        PSE: [PSE36],

        # Entering Long Mode requires that %CR4.PAE is set.  The NX pagetable
        # bit is only representable in the 64bit PTE format offered by PAE.
        PAE: [LM, NX],

        TSC: [TSC_DEADLINE, RDTSCP, TSC_ADJUST, ITSC],

        # APIC is special, but X2APIC does depend on APIC being available in
        # the first place.
        APIC: [X2APIC, TSC_DEADLINE, EXTAPIC],

        # AMD built MMXExtentions and 3DNow as extentions to MMX.
        MMX: [MMXEXT, _3DNOW],

        # The FXSAVE/FXRSTOR instructions were introduced into hardware before
        # SSE, which is why they behave differently based on %CR4.OSFXSAVE and
        # have their own feature bit.  AMD however introduce the Fast FXSR
        # feature as an optimisation.
        FXSR: [FFXSR, SSE],

        # SSE is taken to mean support for the %XMM registers as well as the
        # instructions.  Several further instruction sets are built on core
        # %XMM support, without specific inter-dependencies.  Additionally
        # AMD has a special mis-alignment sub-mode.
        SSE: [SSE2, MISALIGNSSE],

        # SSE2 was re-specified as core instructions for 64bit.  Also ISA
        # extensions dealing with vectors of integers are added here rather
        # than to SSE.
        SSE2: [SSE3, LM, AESNI, PCLMULQDQ, SHA, GFNI],

        # Other SSEn each depend on their predecessor versions.  AMD
        # Lisbon/Magny-Cours processors implemented SSE4A without SSSE3.
        SSE3: [SSSE3, SSE4A],
        SSSE3: [SSE4_1],
        SSE4_1: [SSE4_2],

        # AMD specify no relationship between POPCNT and SSE4.2.  Intel
        # document that SSE4.2 should be checked for before checking for
        # POPCNT.  However, it has its own feature bit, and operates on GPRs
        # rather than %XMM state, so doesn't inherently depend on SSE.
        # Therefore, we do not specify a dependency between SSE4_2 and POPCNT.
        #
        # SSE4_2: [POPCNT]

        # XSAVE is an extra set of instructions for state management, but
        # doesn't constitue new state itself.  Some of the dependent features
        # are instructions built on top of base XSAVE, while others are new
        # instruction groups which are specified to require XSAVE for state
        # management.
        XSAVE: [XSAVEOPT, XSAVEC, XGETBV1, XSAVES,
                AVX, MPX, PKU, AMX_TILE, LWP],

        # AVX is taken to mean hardware support for 256bit registers (which in
        # practice depends on the VEX prefix to encode), and the instructions
        # themselves.
        #
        # AVX is not taken to mean support for the VEX prefix itself (nor XOP
        # for the XOP prefix).  VEX/XOP-encoded GPR instructions, such as
        # those from the BMI{1,2}, TBM and LWP sets function fine in the
        # absence of any enabled xstate.
        AVX: [FMA, FMA4, F16C, AVX2, XOP, AVX_NE_CONVERT, SM3],

        # This dependency exists solely for the shadow pagetable code.  If the
        # host doesn't have NX support, the shadow pagetable code can't handle
        # SMAP correctly for guests.
        NX: [SMAP],

        # CX16 is only encodable in Long Mode.  LAHF_LM indicates that the
        # SAHF/LAHF instructions are reintroduced in Long Mode.  1GB
        # superpages, PCID and PKU are only available in 4 level paging.
        # NO_LMSL indicates the absense of Long Mode Segment Limits, which
        # have been dropped in hardware.
        LM: [CX16, PCID, LAHF_LM, PAGE1GB, PKU, NO_LMSL, AMX_TILE, CMPCCXADD],

        # AMD K6-2+ and K6-III processors shipped with 3DNow+, beyond the
        # standard 3DNow in the earlier K6 processors.
        _3DNOW: [_3DNOWEXT],

        # This is just the dependency between AVX512 and AVX2 of XSTATE
        # feature flags.  If want to use AVX512, AVX2 must be supported and
        # enabled.  Certain later extensions, acting on 256-bit vectors of
        # integers, better depend on AVX2 than AVX.
        AVX2: [AVX512F, VAES, VPCLMULQDQ, AVX_VNNI, AVX_IFMA, AVX_VNNI_INT8,
               AVX_VNNI_INT16, SHA512, SM4],

        # AVX512F is taken to mean hardware support for 512bit registers
        # (which in practice depends on the EVEX prefix to encode) as well
        # as mask registers, and the instructions themselves. All further
        # AVX512 features are built on top of AVX512F
        AVX512F: [AVX512DQ, AVX512_IFMA, AVX512PF, AVX512ER, AVX512CD,
                  AVX512BW, AVX512VL, AVX512_4VNNIW, AVX512_4FMAPS,
                  AVX512_VNNI, AVX512_VPOPCNTDQ, AVX512_VP2INTERSECT],

        # AVX512 extensions acting on vectors of bytes/words are made
        # dependents of AVX512BW (as to requiring wider than 16-bit mask
        # registers), despite the SDM not formally making this connection.
        AVX512BW: [AVX512_VBMI, AVX512_VBMI2, AVX512_BITALG, AVX512_BF16,
                   AVX512_FP16],

        # Extensions with VEX/EVEX encodings keyed to a separate feature
        # flag are made dependents of their respective legacy feature.
        PCLMULQDQ: [VPCLMULQDQ],
        AESNI: [VAES],

        # The features:
        #   * Single Thread Indirect Branch Predictors
        #   * Speculative Store Bypass Disable
        #   * Predictive Store Forward Disable
        #
        # enumerate new bits in MSR_SPEC_CTRL, and technically enumerate
        # MSR_SPEC_CTRL itself.  AMD further enumerates hints to guide OS
        # behaviour.
        #
        # However, no real hardware will exist with e.g. SSBD but not
        # IBRSB/IBRS, and we pass this MSR directly to guests.  Treating them
        # as dependent features simplifies Xen's logic, and prevents the guest
        # from seeing implausible configurations.
        IBRSB: [STIBP, SSBD, INTEL_PSFD, EIBRS,
                IPRED_CTRL, RRSBA_CTRL, BHI_CTRL],
        IBRS: [AMD_STIBP, AMD_SSBD, PSFD, AUTO_IBRS,
               IBRS_ALWAYS, IBRS_FAST, IBRS_SAME_MODE],
        IBPB: [IBPB_RET, SBPB, IBPB_BRTYPE],
        AMD_STIBP: [STIBP_ALWAYS],

        # In principle the TSXLDTRK insns could also be considered independent.
        RTM: [TSXLDTRK],

        # Enhanced Predictive Store-Forwarding is a informational note on top
        # of PSF.
        PSFD: [EPSF],

        # The ARCH_CAPS CPUID bit enumerates the availability of the whole register.
        ARCH_CAPS: list(range(RDCL_NO, RDCL_NO + 64)),

        # The behaviour described by RRSBA depend on eIBRS being active.
        EIBRS: [RRSBA],

        # AMX-TILE means hardware support for tile registers and general non-
        # computational instructions.  All further AMX features are built on top
        # of AMX-TILE.
        AMX_TILE: [AMX_BF16, AMX_INT8, AMX_FP16, AMX_COMPLEX],
    }

    deep_features = tuple(sorted(deps.keys()))
    state.deep_deps = {}

    for feat in deep_features:

        seen = [feat]
        to_process = list(deps[feat])

        while len(to_process):

            # To debug, uncomment the following lines:
            # def repl(l):
            #     return "[" + ", ".join((state.names[x] for x in l)) + "]"
            # sys.stderr.write("Feature %s, seen %s, to_process %s \n" % \
            #     (state.names[feat], repl(seen), repl(to_process)))

            f = to_process.pop(0)

            if f in seen:
                raise Fail("ERROR: Cycle found with %s when processing %s"
                           % (state.names[f], state.names[feat]))

            seen.append(f)
            to_process = list(set(to_process + deps.get(f, [])))

        state.deep_deps[feat] = seen[1:]

    state.deep_features = deps.keys()
    state.nr_deep_deps = len(state.deep_deps.keys())

    # Calculate the bitfield name declarations.  Leave 4 placeholders on the end
    for word in range(state.nr_entries + 4):

        names = []
        for bit in range(32):

            name = state.names.get(word * 32 + bit, "")

            # Prepend an underscore if the name starts with a digit.
            if name and name[0] in "0123456789":
                name = "_" + name

            # Don't generate names for features fast-forwarded from other
            # state
            if name in ("APIC", "OSXSAVE", "OSPKE"):
                name = ""

            names.append(name.lower())

        if any(names):
            state.bitfields.append("bool " + ":1, ".join(names) + ":1")
        else:
            state.bitfields.append("uint32_t _placeholder_%u" % (word, ))


def write_results(state):
    state.output.write(
"""/*
 * Automatically generated by %s - Do not edit!
 * Source data: %s
 */
#ifndef __XEN_X86__FEATURESET_DATA__
#define __XEN_X86__FEATURESET_DATA__
""" % (sys.argv[0], state.source))

    state.output.write(
"""
#define FEATURESET_NR_ENTRIES %s

#define CPUID_COMMON_1D_FEATURES %s

#define INIT_KNOWN_FEATURES { \\\n%s\n}

#define INIT_SPECIAL_FEATURES { \\\n%s\n}

#define INIT_SIMPLE_OR { \\\n%s\n}

#define INIT_PV_DEF_FEATURES { \\\n%s\n}

#define INIT_PV_MAX_FEATURES { \\\n%s\n}

#define INIT_HVM_SHADOW_DEF_FEATURES { \\\n%s\n}

#define INIT_HVM_SHADOW_MAX_FEATURES { \\\n%s\n}

#define INIT_HVM_HAP_DEF_FEATURES { \\\n%s\n}

#define INIT_HVM_HAP_MAX_FEATURES { \\\n%s\n}

#define NR_DEEP_DEPS %sU

#define INIT_DEEP_FEATURES { \\\n%s\n}

#define INIT_DEEP_DEPS { \\
""" % (state.nr_entries,
       next(featureset_to_uint32s(state.common_1d, 1)),
       format_uint32s(state, state.names.keys(), 4),
       format_uint32s(state, state.raw['!'], 4),
       format_uint32s(state, state.raw['|'], 4),
       format_uint32s(state, state.pv_def, 4),
       format_uint32s(state, state.pv_max, 4),
       format_uint32s(state, state.hvm_shadow_def, 4),
       format_uint32s(state, state.hvm_shadow_max, 4),
       format_uint32s(state, state.hvm_hap_def, 4),
       format_uint32s(state, state.hvm_hap_max, 4),
       state.nr_deep_deps,
       format_uint32s(state, state.deep_features, 4),
       ))

    for dep in sorted(state.deep_deps.keys()):
        state.output.write(
            "    { %#xU, /* %s */ { \\\n%s\n    }, }, \\\n"
            % (dep, state.names[dep],
               format_uint32s(state, state.deep_deps[dep], 8)
           ))

    state.output.write(
"""}

#define INIT_FEATURE_NAME_TO_VAL { \\
""")

    for name, bit in sorted(state.values.items()):
        state.output.write(
            '    { "%s", %sU },\\\n' % (name, bit)
            )

    state.output.write(
"""}

""")

    state.output.write(
"""
#define INIT_FEATURE_VAL_TO_NAME { \\
""")

    for name, bit in sorted(state.values.items()):
        state.output.write(
            '    [%s] = "%s",\\\n' % (bit, name)
            )

        # Add the other alias for 1d/e1d common bits.  64 is the difference
        # between 1d and e1d.
        if bit in state.common_1d:
            state.output.write(
                '    [%s] = "%s",\\\n' % (64 + bit, name)
            )

    # Pad to an exact multiple of FEATURESET_SIZE if necessary
    pad_feat = state.nr_entries * 32 - 1
    if not state.names.get(pad_feat):
        state.output.write(
            '    [%s] = NULL,\\\n' % (pad_feat, )
        )

    state.output.write(
"""}

""")

    for idx, text in enumerate(state.bitfields):
        state.output.write(
            "#define CPUID_BITFIELD_%d \\\n    %s\n\n"
            % (idx, text))

    state.output.write(
"""
#endif /* __XEN_X86__FEATURESET_DATA__ */
""")


def open_file_or_fd(val, mode, buffering):
    """
    If 'val' looks like a decimal integer, open it as an fd.  If not, try to
    open it as a regular file.
    """

    fd = -1
    try:
        # Does it look like an integer?
        try:
            fd = int(val, 10)
        except ValueError:
            pass

        if fd == 0:
            return sys.stdin
        elif fd == 1:
            return sys.stdout
        elif fd == 2:
            return sys.stderr

        # Try to open it...
        if fd != -1:
            return os.fdopen(fd, mode, buffering)
        else:
            return open(val, mode, buffering)

    except StandardError:
        e = sys.exc_info()[1]
        if fd != -1:
            raise Fail("Unable to open fd %d: %s: %s" %
                       (fd, e.__class__.__name__, e))
        else:
            raise Fail("Unable to open file '%s': %s: %s" %
                       (val, e.__class__.__name__, e))

    raise SystemExit(2)

def main():
    from optparse import OptionParser

    # Change stdout to be line-buffered.
    sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 1)

    parser = OptionParser(usage = "%prog [options] -i INPUT -o OUTPUT",
                          description =
                          "Process featureset information")

    parser.add_option("-i", "--in", dest = "fin", metavar = "<FD or FILE>",
                      default = "0",
                      help = "Featureset definitions")
    parser.add_option("-o", "--out", dest = "fout", metavar = "<FD or FILE>",
                      default = "1",
                      help = "Featureset calculated information")

    opts, _ = parser.parse_args()

    if opts.fin is None or opts.fout is None:
        parser.print_help(sys.stderr)
        raise SystemExit(1)

    state = State(opts.fin, opts.fout)

    parse_definitions(state)
    crunch_numbers(state)
    write_results(state)


if __name__ == "__main__":
    try:
        sys.exit(main())
    except Fail:
        e = sys.exc_info()[1]
        sys.stderr.write("%s: Fail: %s\n" %
                         (os.path.abspath(sys.argv[0]), str(e)))
        sys.exit(1)
    except SystemExit:
        e = sys.exc_info()[1]
        sys.exit(e.code)
    except KeyboardInterrupt:
        sys.exit(2)



#!/usr/bin/python

from __future__ import print_function

import os
import sys

try:
    sys.path.append(os.environ['LIBXL_SRC_DIR'])
except:
    # If we get here, then we expect the 'import idl'
    # expression to fail. That error is more informative,
    # so let it happen.
    pass

import idl

# Go versions of some builtin types.
# Append the libxl-defined builtins after IDL parsing.
builtin_type_names = {
    idl.bool.typename: 'bool',
    idl.string.typename: 'string',
    idl.integer.typename: 'int',
    idl.uint8.typename: 'byte',
    idl.uint16.typename: 'uint16',
    idl.uint32.typename: 'uint32',
    idl.uint64.typename: 'uint64',
}

# Some go keywords that conflict with field names in libxl structs.
go_keywords = ['type', 'func']

go_builtin_types = ['bool', 'string', 'int', 'byte',
                    'uint16', 'uint32', 'uint64']

# cgo preamble for xenlight_helpers.go, created during type generation and
# written later.
cgo_helpers_preamble = []

def xenlight_golang_generate_types(path = None, types = None, comment = None):
    """
    Generate a .go file (types.gen.go by default)
    that contains a Go type for each type in types.
    """
    if path is None:
        path = 'types.gen.go'

    with open(path, 'w') as f:
        if comment is not None:
            f.write(comment)
        f.write('package xenlight\n\n')

        for ty in types:
            (tdef, extras) = xenlight_golang_type_define(ty)

            f.write(tdef)
            f.write('\n')

            # Append extra types
            for extra in extras:
                f.write(extra)
                f.write('\n')

def xenlight_golang_type_define(ty = None):
    """
    Generate the Go type definition of ty.

    Return a tuple that contains a string with the
    type definition, and a (potentially empty) list
    of extra definitions that are associated with
    this type.
    """
    if isinstance(ty, idl.Enumeration):
        return (xenlight_golang_define_enum(ty), [])

    elif isinstance(ty, idl.Aggregate):
        return xenlight_golang_define_struct(ty)

def xenlight_golang_define_enum(ty = None):
    s = ''
    typename = ''

    if ty.typename is not None:
        typename = xenlight_golang_fmt_name(ty.typename)
        s += 'type {0} int\n'.format(typename)

    # Start const block
    s += 'const(\n'

    for v in ty.values:
        name = xenlight_golang_fmt_name(v.name)
        s += '{0} {1} = {2}\n'.format(name, typename, v.value)

    # End const block
    s += ')\n'

    return s

def xenlight_golang_define_struct(ty = None, typename = None, nested = False):
    s = ''
    extras = []
    name = ''

    if typename is not None:
        name = xenlight_golang_fmt_name(typename)
    else:
        name = xenlight_golang_fmt_name(ty.typename)

    # Begin struct definition
    if nested:
        s += '{0} struct {{\n'.format(name)
    else:
        s += 'type {0} struct {{\n'.format(name)

    # Write struct fields
    for f in ty.fields:
        if f.type.typename is not None:
            if isinstance(f.type, idl.Array):
                typename = f.type.elem_type.typename
                typename = xenlight_golang_fmt_name(typename)
                name     = xenlight_golang_fmt_name(f.name)

                s += '{0} []{1}\n'.format(name, typename)
            else:
                typename = f.type.typename
                typename = xenlight_golang_fmt_name(typename)
                name     = xenlight_golang_fmt_name(f.name)

                s += '{0} {1}\n'.format(name, typename)

        elif isinstance(f.type, idl.Struct):
            r = xenlight_golang_define_struct(f.type, typename=f.name, nested=True)

            s += r[0]
            extras.extend(r[1])

        elif isinstance(f.type, idl.KeyedUnion):
            r = xenlight_golang_define_union(f.type, ty.typename, f.name)

            s += r[0]
            extras.extend(r[1])

        else:
            raise Exception('type {0} not supported'.format(f.type))

    # End struct definition
    s += '}\n'

    return (s,extras)

def xenlight_golang_define_union(ty = None, struct_name = '', union_name = ''):
    """
    Generate the Go translation of a KeyedUnion.

    Define an unexported interface to be used as
    the type of the union. Then, define a struct
    for each field of the union which implements
    that interface.
    """
    s = ''
    extras = []

    interface_name = '{0}_{1}_union'.format(struct_name, ty.keyvar.name)
    interface_name = xenlight_golang_fmt_name(interface_name)

    s += 'type {0} interface {{\n'.format(interface_name)
    s += 'is{0}()\n'.format(interface_name)
    s += '}\n'

    extras.append(s)

    for f in ty.fields:
        if f.type is None:
            continue

        # Define struct
        name = '{0}_{1}_union_{2}'.format(struct_name, ty.keyvar.name, f.name)
        r = xenlight_golang_define_struct(f.type, typename=name)
        extras.append(r[0])
        extras.extend(r[1])

        # This typeof trick ensures that the fields used in the cgo struct
        # used for marshaling are the same as the fields of the union in the
        # actual C type, and avoids re-defining all of those fields.
        s = 'typedef typeof(((struct {0} *)NULL)->{1}.{2}){3};'
        s = s.format(struct_name, union_name, f.name, name)
        cgo_helpers_preamble.append(s)

        # Define function to implement 'union' interface
        name = xenlight_golang_fmt_name(name)
        s = 'func (x {0}) is{1}(){{}}\n'.format(name, interface_name)
        extras.append(s)

    fname = xenlight_golang_fmt_name(ty.keyvar.name)
    ftype = xenlight_golang_fmt_name(ty.keyvar.type.typename)
    s = '{0} {1}\n'.format(fname, ftype)

    fname = xenlight_golang_fmt_name('{0}_union'.format(ty.keyvar.name))
    s += '{0} {1}\n'.format(fname, interface_name)

    return (s,extras)

def xenlight_golang_generate_helpers(path = None, types = None, comment = None):
    """
    Generate a .go file (helpers.gen.go by default)
    that contains helper functions for marshaling between
    C and Go types.
    """
    if path is None:
        path = 'helpers.gen.go'

    with open(path, 'w') as f:
        if comment is not None:
            f.write(comment)
        f.write('package xenlight\n\n')
        f.write('import (\n"unsafe"\n"errors"\n"fmt"\n)\n')

        # Cgo preamble
        f.write('/*\n')
        f.write('#cgo LDFLAGS: -lxenlight\n')
        f.write('#include <stdlib.h>\n')
        f.write('#include <libxl.h>\n')
        f.write('\n')

        for s in cgo_helpers_preamble:
            f.write(s)
            f.write('\n')

        f.write('*/\nimport "C"\n')

        for ty in types:
            if not isinstance(ty, idl.Struct):
                continue

            f.write(xenlight_golang_define_constructor(ty))
            f.write('\n')

            (fdef, extras) = xenlight_golang_define_from_C(ty)

            f.write(fdef)
            f.write('\n')

            for extra in extras:
                f.write(extra)
                f.write('\n')

            f.write(xenlight_golang_define_to_C(ty))
            f.write('\n')

def xenlight_golang_define_from_C(ty = None):
    """
    Define the fromC marshaling function for the type
    represented by ty.
    """
    func = 'func (x *{0}) fromC(xc *C.{1}) error {{\n {2}\n return nil}}\n'

    goname = xenlight_golang_fmt_name(ty.typename)
    cname  = ty.typename

    body = ''
    extras = []

    for f in ty.fields:
        if f.type.typename is not None:
            if isinstance(f.type, idl.Array):
                body += xenlight_golang_array_from_C(f)
                continue

            body += xenlight_golang_convert_from_C(f)

        elif isinstance(f.type, idl.Struct):
            # Go through the fields of the anonymous nested struct.
            for nf in f.type.fields:
                body += xenlight_golang_convert_from_C(nf,outer_name=f.name)

        elif isinstance(f.type, idl.KeyedUnion):
            r = xenlight_golang_union_from_C(f.type, f.name, ty.typename)

            body += r[0]
            extras.extend(r[1])

        else:
            raise Exception('type {0} not supported'.format(f.type))

    return (func.format(goname, cname, body), extras)

def xenlight_golang_convert_from_C(ty = None, outer_name = None, cvarname = None):
    """
    Returns a line of Go code that converts the C type represented
    by ty to its corresponding Go type.

    If outer_name is set, the type is treated as nested within another field
    named outer_name.
    """
    s = ''

    # Use 'xc' as the name for the C variable unless otherwise specified
    if cvarname is None:
        cvarname = 'xc'

    gotypename = xenlight_golang_fmt_name(ty.type.typename)
    goname     = xenlight_golang_fmt_name(ty.name)
    cname      = ty.name

    # In cgo, C names that conflict with Go keywords can be
    # accessed by prepending an underscore to the name.
    if cname in go_keywords:
        cname = '_' + cname

    # If outer_name is set, treat this as nested.
    if outer_name is not None:
        goname = '{0}.{1}'.format(xenlight_golang_fmt_name(outer_name), goname)
        cname  = '{0}.{1}'.format(outer_name, cname)

    # Types that satisfy this condition can be easily casted or
    # converted to a Go builtin type.
    is_castable = (ty.type.json_parse_type == 'JSON_INTEGER' or
                   isinstance(ty.type, idl.Enumeration) or
                   gotypename in go_builtin_types)

    if not is_castable:
        # If the type is not castable, we need to call its fromC
        # function.
        s += 'if err := x.{0}.fromC(&{1}.{2});'.format(goname,cvarname,cname)
        s += 'err != nil {{\nreturn fmt.Errorf("converting field {0}: %v", err)\n}}\n'.format(goname)

    elif gotypename == 'string':
        # Use the cgo helper for converting C strings.
        s += 'x.{0} = C.GoString({1}.{2})\n'.format(goname,cvarname,cname)

    else:
        s += 'x.{0} = {1}({2}.{3})\n'.format(goname,gotypename,cvarname,cname)

    return s

def xenlight_golang_union_from_C(ty = None, union_name = '', struct_name = ''):
    extras = []

    keyname   = ty.keyvar.name
    gokeyname = xenlight_golang_fmt_name(keyname)
    keytype   = ty.keyvar.type.typename
    gokeytype = xenlight_golang_fmt_name(keytype)
    field_name = xenlight_golang_fmt_name('{0}_union'.format(keyname))

    interface_name = '{0}_{1}_union'.format(struct_name, keyname)
    interface_name = xenlight_golang_fmt_name(interface_name)

    cgo_keyname = keyname
    if cgo_keyname in go_keywords:
        cgo_keyname = '_' + cgo_keyname

    cases = {}

    for f in ty.fields:
        val = '{0}_{1}'.format(keytype, f.name)
        val = xenlight_golang_fmt_name(val)

        # Add to list of cases to make for the switch
        # statement below.
        cases[f.name] = (val, f.type)

        if f.type is None:
            continue

        # Define fromC func for 'union' struct.
        typename   = '{0}_{1}_union_{2}'.format(struct_name,keyname,f.name)
        gotypename = xenlight_golang_fmt_name(typename)

        # Define the function here. The cases for keyed unions are a little
        # different.
        s = 'func (x *{0}) fromC(xc *C.{1}) error {{\n'.format(gotypename,struct_name)
        s += 'if {0}(xc.{1}) != {2} {{\n'.format(gokeytype,cgo_keyname,val)
        err_string = '"expected union key {0}"'.format(val)
        s += 'return errors.New({0})\n'.format(err_string)
        s += '}\n\n'
        s += 'tmp := (*C.{0})(unsafe.Pointer(&xc.{1}[0]))\n'.format(typename,union_name)

        for nf in f.type.fields:
            if isinstance(nf.type, idl.Array):
                s += xenlight_golang_array_from_C(nf,cvarname='tmp')
                continue

            s += xenlight_golang_convert_from_C(nf,cvarname='tmp')

        s += 'return nil\n'
        s += '}\n'

        extras.append(s)

    s = 'x.{0} = {1}(xc.{2})\n'.format(gokeyname,gokeytype,cgo_keyname)
    s += 'switch x.{0}{{\n'.format(gokeyname)

    # Create switch statement to determine which 'union element'
    # to populate in the Go struct.
    for case_name, case_tuple in sorted(cases.items()):
        (case_val, case_type) = case_tuple

        s += 'case {0}:\n'.format(case_val)

        if case_type is None:
            s += "x.{0} = nil\n".format(field_name)
            continue

        gotype = '{0}_{1}_union_{2}'.format(struct_name,keyname,case_name)
        gotype = xenlight_golang_fmt_name(gotype)
        goname = '{0}_{1}'.format(keyname,case_name)
        goname = xenlight_golang_fmt_name(goname,exported=False)

        s += 'var {0} {1}\n'.format(goname, gotype)
        s += 'if err := {0}.fromC(xc);'.format(goname)
        s += 'err != nil {{\n return fmt.Errorf("converting field {0}: %v", err)\n}}\n'.format(goname)

        s += 'x.{0} = &{1}\n'.format(field_name, goname)

    # End switch statement
    s += 'default:\n'
    err_string = '"invalid union key \'%v\'", x.{0}'.format(gokeyname)
    s += 'return fmt.Errorf({0})'.format(err_string)
    s += '}\n'

    return (s,extras)

def xenlight_golang_array_from_C(ty = None, cvarname = 'xc'):
    """
    Convert C array to Go slice using the method
    described here:

    https://github.com/golang/go/wiki/cgo#turning-c-arrays-into-go-slices
    """
    s = ''

    gotypename = xenlight_golang_fmt_name(ty.type.elem_type.typename)
    goname     = xenlight_golang_fmt_name(ty.name)
    ctypename  = ty.type.elem_type.typename
    cname      = ty.name
    cslice     = 'c{0}'.format(goname)
    clenvar    = ty.type.lenvar.name

    s += 'x.{0} = nil\n'.format(goname)
    s += 'if n := int({0}.{1}); n > 0 {{\n'.format(cvarname,clenvar)
    s += '{0} := '.format(cslice)
    s +='(*[1<<28]C.{0})(unsafe.Pointer({1}.{2}))[:n:n]\n'.format(ctypename, cvarname, cname)
    s += 'x.{0} = make([]{1}, n)\n'.format(goname, gotypename)
    s += 'for i, v := range {0} {{\n'.format(cslice)

    is_enum = isinstance(ty.type.elem_type,idl.Enumeration)
    if gotypename in go_builtin_types or is_enum:
        s += 'x.{0}[i] = {1}(v)\n'.format(goname, gotypename)
    else:
        s += 'if err := x.{0}[i].fromC(&v); err != nil {{\n'.format(goname)
        s += 'return fmt.Errorf("converting field {0}: %v", err) }}\n'.format(goname)

    s += '}\n}\n'

    return s

def xenlight_golang_define_to_C(ty = None, typename = None, nested = False):
    """
    Define the toC marshaling function for the type
    represented by ty.
    """
    func = 'func (x *{0}) toC(xc *C.{1}) (err error){{{2}\n return nil\n }}\n'
    body = ''

    if ty.dispose_fn is not None:
        body += 'defer func(){{\nif err != nil{{\nC.{0}(xc)}}\n}}()\n\n'.format(ty.dispose_fn)

    goname = xenlight_golang_fmt_name(ty.typename)
    cname  = ty.typename

    for f in ty.fields:
        if f.type.typename is not None:
            if isinstance(f.type, idl.Array):
                body += xenlight_golang_array_to_C(f)
                continue

            body += xenlight_golang_convert_to_C(f)

        elif isinstance(f.type, idl.Struct):
            for nf in f.type.fields:
                body += xenlight_golang_convert_to_C(nf, outer_name=f.name)

        elif isinstance(f.type, idl.KeyedUnion):
            body += xenlight_golang_union_to_C(f.type, f.name, ty.typename)

        else:
            raise Exception('type {0} not supported'.format(f.type))

    return func.format(goname, cname, body)

def xenlight_golang_convert_to_C(ty = None, outer_name = None,
                                 govarname = None, cvarname = None):
    """
    Returns a line of Go code that converts the Go type represented
    by ty to its corresponding Go type.

    If outer_name is set, the type is treated as nested within another field
    named outer_name.
    """
    s = ''

    # Use 'xc' as the name for the C variable unless otherwise specified.
    if cvarname is None:
        cvarname = 'xc'

    # Use 'x' as the name for the Go variable unless otherwise specified.
    if govarname is None:
        govarname = 'x'

    gotypename = xenlight_golang_fmt_name(ty.type.typename)
    ctypename  = ty.type.typename
    goname     = xenlight_golang_fmt_name(ty.name)
    cname      = ty.name

    # In cgo, C names that conflict with Go keywords can be
    # accessed by prepending an underscore to the name.
    if cname in go_keywords:
        cname = '_' + cname

    # If outer_name is set, treat this as nested.
    if outer_name is not None:
        goname = '{0}.{1}'.format(xenlight_golang_fmt_name(outer_name), goname)
        cname  = '{0}.{1}'.format(outer_name, cname)

    is_castable = (ty.type.json_parse_type == 'JSON_INTEGER' or
                   isinstance(ty.type, idl.Enumeration) or
                   gotypename in go_builtin_types)

    if not is_castable:
        s += 'if err := {0}.{1}.toC(&{2}.{3}); err != nil {{\n'.format(govarname,goname,
                                                                   cvarname,cname)
        s += 'return fmt.Errorf("converting field {0}: %v", err)\n}}\n'.format(goname)

    elif gotypename == 'string':
        # Use the cgo helper for converting C strings.
        s += 'if {0}.{1} != "" {{\n'.format(govarname,goname)
        s += '{0}.{1} = C.CString({2}.{3})}}\n'.format(cvarname,cname,
                                                   govarname,goname)

    else:
        s += '{0}.{1} = C.{2}({3}.{4})\n'.format(cvarname,cname,ctypename,
                                            govarname,goname)

    return s

def xenlight_golang_union_to_C(ty = None, union_name = '',
                               struct_name = ''):
    keyname   = ty.keyvar.name
    gokeyname = xenlight_golang_fmt_name(keyname)
    keytype   = ty.keyvar.type.typename
    gokeytype = xenlight_golang_fmt_name(keytype)

    interface_name = '{0}_{1}_union'.format(struct_name, keyname)
    interface_name = xenlight_golang_fmt_name(interface_name)

    cgo_keyname = keyname
    if cgo_keyname in go_keywords:
        cgo_keyname = '_' + cgo_keyname


    s = 'xc.{0} = C.{1}(x.{2})\n'.format(cgo_keyname,keytype,gokeyname)
    s += 'switch x.{0}{{\n'.format(gokeyname)

    # Create switch statement to determine how to populate the C union.
    for f in ty.fields:
        key_val = '{0}_{1}'.format(keytype, f.name)
        key_val = xenlight_golang_fmt_name(key_val)

        s += 'case {0}:\n'.format(key_val)

        if f.type is None:
            s += "break\n"
            continue

        cgotype = '{0}_{1}_union_{2}'.format(struct_name,keyname,f.name)
        gotype  = xenlight_golang_fmt_name(cgotype)

        field_name = xenlight_golang_fmt_name('{0}_union'.format(keyname))
        s += 'tmp, ok := x.{0}.(*{1})\n'.format(field_name,gotype)
        s += 'if !ok {\n'
        s += 'return errors.New("wrong type for union key {0}")\n'.format(keyname)
        s += '}\n'

        s += 'var {0} C.{1}\n'.format(f.name,cgotype)
        for uf in f.type.fields:
            if isinstance(uf.type, idl.Array):
                s += xenlight_golang_array_to_C(uf, cvarname=f.name,
                                                govarname="tmp")
                continue

            s += xenlight_golang_convert_to_C(uf,cvarname=f.name,
                                              govarname='tmp')

        # The union is still represented as Go []byte.
        s += '{0}Bytes := C.GoBytes(unsafe.Pointer(&{1}),C.sizeof_{2})\n'.format(f.name,
                                                                              f.name,
                                                                              cgotype)
        s += 'copy(xc.{0}[:],{1}Bytes)\n'.format(union_name,f.name)

    # End switch statement
    s += 'default:\n'
    err_string = '"invalid union key \'%v\'", x.{0}'.format(gokeyname)
    s += 'return fmt.Errorf({0})'.format(err_string)
    s += '}\n'

    return s

def xenlight_golang_array_to_C(ty = None, cvarname="xc", govarname="x"):
    s = ''

    gotypename = xenlight_golang_fmt_name(ty.type.elem_type.typename)
    goname     = xenlight_golang_fmt_name(ty.name)
    ctypename  = ty.type.elem_type.typename
    cname      = ty.name
    clenvar    = ty.type.lenvar.name
    golenvar   = xenlight_golang_fmt_name(clenvar,exported=False)

    is_enum = isinstance(ty.type.elem_type,idl.Enumeration)
    if gotypename in go_builtin_types or is_enum:
        s += 'if {0} := len({1}.{2}); {3} > 0 {{\n'.format(golenvar,govarname,goname,golenvar)
        s += '{0}.{1} = (*C.{2})(C.malloc(C.size_t({3}*{4})))\n'.format(cvarname,cname,ctypename,
                                                                   golenvar,golenvar)
        s += '{0}.{1} = C.int({2})\n'.format(cvarname,clenvar,golenvar)
        s += 'c{0} := (*[1<<28]C.{1})(unsafe.Pointer({2}.{3}))[:{4}:{5}]\n'.format(goname,
                                                                      ctypename,cvarname,cname,
                                                                      golenvar,golenvar)
        s += 'for i,v := range {0}.{1} {{\n'.format(govarname,goname)
        s += 'c{0}[i] = C.{1}(v)\n'.format(goname,ctypename)
        s += '}\n}\n'

        return s

    s += 'if {0} := len({1}.{2}); {3} > 0 {{\n'.format(golenvar,govarname,goname,golenvar)
    s += '{0}.{1} = (*C.{2})(C.malloc(C.ulong({3})*C.sizeof_{4}))\n'.format(cvarname,cname,ctypename,
                                                                   golenvar,ctypename)
    s += '{0}.{1} = C.int({2})\n'.format(cvarname,clenvar,golenvar)
    s += 'c{0} := (*[1<<28]C.{1})(unsafe.Pointer({2}.{3}))[:{4}:{5}]\n'.format(goname,
                                                                         ctypename,cvarname,cname,
                                                                         golenvar,golenvar)
    s += 'for i,v := range {0}.{1} {{\n'.format(govarname,goname)
    s += 'if err := v.toC(&c{0}[i]); err != nil {{\n'.format(goname)
    s += 'return fmt.Errorf("converting field {0}: %v", err)\n'.format(goname)
    s += '}\n}\n}\n'

    return s

def xenlight_golang_define_constructor(ty = None):
    s = ''

    ctypename  = ty.typename
    gotypename = xenlight_golang_fmt_name(ctypename)

    # Since this func is exported, add a comment as per Go conventions.
    s += '// New{0} returns an instance of {1}'.format(gotypename,gotypename)
    s += ' initialized with defaults.\n'

    # If a struct has a keyed union, an extra argument is
    # required in the function signature, and an extra _init
    # call is needed.
    params   = []
    init_fns = []

    # Add call to parent init_fn first.
    init_fns.append('C.{0}(&xc)'.format(ty.init_fn))

    for f in ty.fields:
        if not isinstance(f.type, idl.KeyedUnion):
            continue

        param = f.type.keyvar

        param_ctype  = param.type.typename
        param_gotype = xenlight_golang_fmt_name(param_ctype)
        param_goname = xenlight_golang_fmt_name(param.name,exported=False)

        # Serveral keyed unions use 'type' as the key variable name. In
        # that case, prepend the first letter of the Go type name.
        if param_goname == 'type':
            param_goname = '{0}type'.format(param_gotype.lower()[0])

        # Add call to keyed union's init_fn.
        init_fns.append('C.{0}_{1}(&xc, C.{2}({3}))'.format(ty.init_fn,
                                                        param.name,
                                                        param_ctype,
                                                        param_goname))

        # Add to params list.
        params.append('{0} {1}'.format(param_goname, param_gotype))

    # Define function
    s += 'func New{0}({1}) (*{2}, error) {{\n'.format(gotypename,
                                                   ','.join(params),
                                                   gotypename)

    # Declare variables.
    s += 'var (\nx {0}\nxc C.{1})\n\n'.format(gotypename, ctypename)

    # Write init_fn calls.
    s += '\n'.join(init_fns)
    s += '\n'

    # Make sure dispose_fn get's called when constructor
    # returns.
    if ty.dispose_fn is not None:
        s += 'defer C.{0}(&xc)\n'.format(ty.dispose_fn)

    s += '\n'

    # Call fromC to initialize Go type.
    s += 'if err := x.fromC(&xc); err != nil {\n'
    s += 'return nil, err }\n\n'
    s += 'return &x, nil}\n'

    return s

def xenlight_golang_fmt_name(name, exported = True):
    """
    Take a given type name and return an
    appropriate Go type name.
    """
    if name in builtin_type_names.keys():
        return builtin_type_names[name]

    # Name is not a builtin, format it for Go.
    words = name.split('_')

    # Remove 'libxl' prefix
    if words[0].lower() == 'libxl':
        words.remove(words[0])

    if exported:
        return ''.join(x.title() for x in words)

    return words[0] + ''.join(x.title() for x in words[1:])

if __name__ == '__main__':
    if len(sys.argv) != 4:
        print("Usage: gengotypes.py <idl> <types.gen.go> <helpers.gen.go>", file=sys.stderr)
        sys.exit(1)

    idlname = sys.argv[1]
    path_types = sys.argv[2]
    path_helpers = sys.argv[3]

    (builtins, types) = idl.parse(idlname)

    for b in builtins:
        name = b.typename
        builtin_type_names[name] = xenlight_golang_fmt_name(name)

    header_comment="""// Code generated by {}. DO NOT EDIT.
// source: {}

""".format(os.path.basename(sys.argv[0]),
           os.path.basename(sys.argv[1]))

    xenlight_golang_generate_types(types=types,
                                   path=path_types,
                                   comment=header_comment)
    xenlight_golang_generate_helpers(types=types,
                                     path=path_helpers,
                                     comment=header_comment)



# Prefer setuptools, fall back to distutils
try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension
import os, sys

XEN_ROOT = "../.."

SHLIB_libxenctrl = os.environ['SHLIB_libxenctrl'].split()
SHLIB_libxenguest = os.environ['SHLIB_libxenguest'].split()
SHLIB_libxenstore = os.environ['SHLIB_libxenstore'].split()

extra_compile_args  = [ "-fno-strict-aliasing" ]

PATH_XEN      = XEN_ROOT + "/tools/include"
PATH_LIBXENTOOLLOG = XEN_ROOT + "/tools/libs/toollog"
PATH_LIBXENEVTCHN = XEN_ROOT + "/tools/libs/evtchn"
PATH_LIBXENCTRL = XEN_ROOT + "/tools/libs/ctrl"
PATH_LIBXENGUEST = XEN_ROOT + "/tools/libs/guest"
PATH_XENSTORE = XEN_ROOT + "/tools/libs/store"

xc = Extension("xc",
               extra_compile_args = extra_compile_args,
               include_dirs       = [ PATH_XEN,
                                      PATH_LIBXENTOOLLOG + "/include",
                                      PATH_LIBXENEVTCHN + "/include",
                                      PATH_LIBXENCTRL + "/include",
                                      PATH_LIBXENGUEST + "/include",
                                      "xen/lowlevel/xc" ],
               library_dirs       = [ PATH_LIBXENCTRL, PATH_LIBXENGUEST ],
               libraries          = [ "xenctrl", "xenguest" ],
               depends            = [ PATH_LIBXENCTRL + "/libxenctrl.so", PATH_LIBXENGUEST + "/libxenguest.so" ],
               extra_link_args    = SHLIB_libxenctrl + SHLIB_libxenguest,
               sources            = [ "xen/lowlevel/xc/xc.c" ])

xs = Extension("xs",
               extra_compile_args = extra_compile_args,
               include_dirs       = [ PATH_XEN, PATH_XENSTORE + "/include", "xen/lowlevel/xs" ],
               library_dirs       = [ PATH_XENSTORE ],
               libraries          = [ "xenstore" ],
               depends            = [ PATH_XENSTORE + "/libxenstore.so" ],
               extra_link_args    = SHLIB_libxenstore,
               sources            = [ "xen/lowlevel/xs/xs.c" ])

plat = os.uname()[0]
modules = [ xc, xs ]

setup(name            = 'xen',
      version         = '3.0',
      description     = 'Xen',
      packages        = ['xen',
                         'xen.migration',
                         'xen.lowlevel',
                        ],
      ext_package = "xen.lowlevel",
      ext_modules = modules
      )



#!/usr/bin/env python3

import re, json

class TagDatabaseError(Exception):
    pass

# This is the dictionary for the rules that translates to proprietary comments:
#  - cppcheck: /* cppcheck-suppress[id] */
#  - coverity: /* coverity[id] */
#  - eclair:   /* -E> hide id 1 "" */
# Add entries to support more analyzers
tool_syntax = {
    "cppcheck":"cppcheck-suppress[VID]",
    "coverity":"coverity[VID]",
    "eclair":"-E> hide VID 1 \"\""
}

def get_xen_tag_regex(tool):
    return rf'(?P<tag>SAF-(?P<id>\d+)-(?P<type>safe|false-positive-{tool}))'


def get_xen_tag_index_type_regex(tool):
    return rf'^{get_xen_tag_regex(tool)}$'


def get_xen_tag_comment_regex(tool):
    before_comment = r'(?P<before>.*)'
    comment = rf'(?P<comment>/\* +{get_xen_tag_regex(tool)}.*\*/)'
    return rf'^(?P<full_line>{before_comment}{comment})$'


# Returns a data structure containing dictionaries for safe and false-positive-*
# Xen tags, the key is the unique index of the tag and the content is the
# proprietary in-code comment to be used when the tag is found in the codebase
def load_tag_database(tool, input_files, data_struct = None, schema = "safe"):
    ret = data_struct if data_struct is not None else {
        "safe": {},
        "false-positive-" + tool: {}
    }
    database = []

    # Open all input files
    for file in input_files:
        try:
            with open(file, "rt") as handle:
                content = json.load(handle)
                database = database + content['content']
        except json.JSONDecodeError as e:
            raise TagDatabaseError("JSON decoding error in file {}: {}"
                                    .format(file, e))
        except Exception as e:
            raise TagDatabaseError("Can't open file {}: {}"
                                    .format(file, e))

    for entry in database:
        # If the false-positive schema is used, check the proprietary id in the
        # 'violation-id' field, otherwise rely on the "safe" schema.
        if schema == "false-positive":
            proprietary_id = entry['violation-id']
        elif tool in entry['analyser']:
            proprietary_id = entry['analyser'][tool]
        else:
            proprietary_id = ""
        if proprietary_id != "":
            comment=tool_syntax[tool].replace("VID",proprietary_id)
            # Regex to capture the index of the Xen tag and the schema
            xen_tag = re.match(get_xen_tag_index_type_regex(tool), entry["id"])
            if xen_tag and xen_tag.group('id') and xen_tag.group('type'):
                # Save in safe or false-positive-* the key {#id: "comment"}
                id_number = int(xen_tag.group('id'))
                key = xen_tag.group('type')
                ret[key][id_number] = "/* {} */".format(comment)
            else:
                raise TagDatabaseError(
                        "Error in database file, entry {} has unexpected "
                        "format.".format(entry["id"])
                    )

    return ret


def substitute_tags(tool, input_file, grep_struct, subs_rules):
    try:
        with open(grep_struct["file"], "wt") as outfile:

            try:
                with open(input_file, "rt") as infile:
                    parsed_content = infile.readlines()
            except Exception as e:
                raise TagDatabaseError("Issue with reading file {}: {}"
                                       .format(input_file, e))

            # grep_struct contains the line number where the comments are, the
            # line number starts from 1 but in the array the first line is zero.
            # For every line where there is a Xen tag comment, get the Xen tag
            # that is in the capture group zero, extract from the Xen tag the
            # unique index and the type (safe, false-positive-*) and with those
            # information access the subs_rules dictionary to see if there is
            # a match
            for line_number in grep_struct["matches"]:
                xen_tag = grep_struct["matches"][line_number]['tag']
                xen_tag_regex_obj = re.match(get_xen_tag_index_type_regex(tool),
                                             xen_tag)
                id_number = int(xen_tag_regex_obj.group('id'))
                key = xen_tag_regex_obj.group('type')
                if id_number in subs_rules[key]:
                    comment_in = grep_struct["matches"][line_number]['comment']
                    before = grep_struct["matches"][line_number]['before']
                    comment_out = subs_rules[key][id_number]
                    if before != '' and not re.match(r'^[ \t]+$', before):
                        # The comment is at the end of some line with some code
                        if tool == "eclair":
                            # Eclair supports comment at the end of the line, so
                            # the only thing to do is use the right syntax in
                            # the comment, the default version of it is
                            # deviating the current line and the next one
                            comment_out = re.sub(r'\d+ ""', '0 ""', comment_out)
                        else:
                            # Other tool does not support deviating the same
                            # line of the comment, so we use a trick and we use
                            # the comment at the end of the previous line
                            if line_number-2 < 0:
                                raise TagDatabaseError(
                                    "The comment {} using the tool '{}' can't "
                                    "stay at the end of the line 1."
                                    .format(comment_in, tool)
                                )
                            parsed_content[line_number-2] = \
                                parsed_content[line_number-2].replace("\n",
                                    comment_out + '\n')
                            comment_out = ''
                    parsed_content[line_number-1] = re.sub(
                        re.escape(comment_in), comment_out,
                        parsed_content[line_number-1])

            outfile.writelines(parsed_content)
    except Exception as e:
        raise TagDatabaseError("Issue with writing file {}: {}"
                               .format(grep_struct["file"], e))



#!/usr/bin/env python

import re,sys

pats = [
 [ r"^\s*#\s*include\s+", r"__InClUdE__ " ],
 [ r"^\s*#\s*ifdef (XEN_HAVE.*)\s+", r"__IfDeF__ \1" ],
 [ r"^\s*#\s*else /\* (XEN_HAVE.*) \*/\s+", r"__ElSe__" ],
 [ r"^\s*#\s*endif /\* (XEN_HAVE.*) \*/\s+", r"__EnDif__" ],
 [ r"^\s*#\s*define\s+([A-Z_]*_GUEST_HANDLE)", r"#define HIDE_\1" ],
 [ r"^\s*#\s*define\s+([a-z_]*_guest_handle)", r"#define hide_\1" ],
 [ r"^\s*#\s*define\s+(u?int64)_aligned_t\s.*aligned.*", r"typedef \1_T __attribute__((aligned(4))) \1_compat_T;" ],
 [ r"XEN_GUEST_HANDLE", r"COMPAT_HANDLE" ],
];

try:
    xlatf = open(sys.argv[1], 'r')
except IndexError:
    print('missing path to xlat.lst argument')
    sys.exit(1)
for line in xlatf.readlines():
    match = re.subn(r"^\s*\?\s+(\w*)\s.*", r"\1", line.rstrip())
    if match[1]:
        pats.append([ r"(struct|union)\s+(%s|xen_%s)\s+(\w)" % (match[0], match[0]),
                      r"\1 @KeeP@\2 \3" ])
xlatf.close()

for pat in pats:
    pat[0] = re.compile(pat[0])

for line in sys.stdin.readlines():
    if 'DEFINE_XEN_GUEST_HANDLE(long)' in line:
        continue
    for pat in pats:
        line = re.sub(pat[0], pat[1], line)
    print(line.rstrip())



# SPDX-License-Identifier: GPL-2.0
"""
Set random values recursively in nested choices.

Kconfig can create a choice-in-choice structure by using 'if' statement.
randconfig should correctly set random choice values.

Related Linux commit: 3b9a19e08960e5cdad5253998637653e592a3c29
"""


def test(conf):
    for i in range(20):
        assert conf.randconfig() == 0
        assert (conf.config_contains('expected_stdout0') or
                conf.config_contains('expected_stdout1') or
                conf.config_contains('expected_stdout2'))



# -*- coding: utf-8 -*-

"""
Legacy migration stream information.

Documentation and record structures for legacy migration, for both libxc
and libxl.
"""

"""
Libxc:

SAVE/RESTORE/MIGRATE PROTOCOL
=============================

The general form of a stream of chunks is a header followed by a
body consisting of a variable number of chunks (terminated by a
chunk with type 0) followed by a trailer.

For a rolling/checkpoint (e.g. remus) migration then the body and
trailer phases can be repeated until an external event
(e.g. failure) causes the process to terminate and commit to the
most recent complete checkpoint.

HEADER
------

unsigned long        : p2m_size

extended-info (PV-only, optional):

  If first unsigned long == ~0UL then extended info is present,
  otherwise unsigned long is part of p2m. Note that p2m_size above
  does not include the length of the extended info.

  extended-info:

    unsigned long    : signature == ~0UL
    uint32_t	        : number of bytes remaining in extended-info

    1 or more extended-info blocks of form:
    char[4]          : block identifier
    uint32_t         : block data size
    bytes            : block data

    defined extended-info blocks:
    "vcpu"		: VCPU context info containing vcpu_guest_context_t.
                       The precise variant of the context structure
                       (e.g. 32 vs 64 bit) is distinguished by
                       the block size.
    "extv"           : Presence indicates use of extended VCPU context in
                       tail, data size is 0.

p2m (PV-only):

  consists of p2m_size bytes comprising an array of xen_pfn_t sized entries.

BODY PHASE - Format A (for live migration or Remus without compression)
----------

A series of chunks with a common header:
  int              : chunk type

If the chunk type is +ve then chunk contains guest memory data, and the
type contains the number of pages in the batch:

    unsigned long[]  : PFN array, length == number of pages in batch
                       Each entry consists of XEN_DOMCTL_PFINFO_*
                       in bits 31-28 and the PFN number in bits 27-0.
    page data        : PAGE_SIZE bytes for each page marked present in PFN
                       array

If the chunk type is -ve then chunk consists of one of a number of
metadata types.  See definitions of XC_SAVE_ID_* below.

If chunk type is 0 then body phase is complete.


BODY PHASE - Format B (for Remus with compression)
----------

A series of chunks with a common header:
  int              : chunk type

If the chunk type is +ve then chunk contains array of PFNs corresponding
to guest memory and type contains the number of PFNs in the batch:

    unsigned long[]  : PFN array, length == number of pages in batch
                       Each entry consists of XEN_DOMCTL_PFINFO_*
                       in bits 31-28 and the PFN number in bits 27-0.

If the chunk type is -ve then chunk consists of one of a number of
metadata types.  See definitions of XC_SAVE_ID_* below.

If the chunk type is -ve and equals XC_SAVE_ID_COMPRESSED_DATA, then the
chunk consists of compressed page data, in the following format:

    unsigned long        : Size of the compressed chunk to follow
    compressed data :      variable length data of size indicated above.
                           This chunk consists of compressed page data.
                           The number of pages in one chunk depends on
                           the amount of space available in the sender's
                           output buffer.

Format of compressed data:
  compressed_data = <deltas>*
  delta           = <marker, run*>
  marker          = (RUNFLAG|SKIPFLAG) bitwise-or RUNLEN [1 byte marker]
  RUNFLAG         = 0
  SKIPFLAG        = 1 << 7
  RUNLEN          = 7-bit unsigned value indicating number of WORDS in the run
  run             = string of bytes of length sizeof(WORD) * RUNLEN

   If marker contains RUNFLAG, then RUNLEN * sizeof(WORD) bytes of data following
  the marker is copied into the target page at the appropriate offset indicated by
  the offset_ptr
   If marker contains SKIPFLAG, then the offset_ptr is advanced
  by RUNLEN * sizeof(WORD).

If chunk type is 0 then body phase is complete.

There can be one or more chunks with type XC_SAVE_ID_COMPRESSED_DATA,
containing compressed pages. The compressed chunks are collated to form
one single compressed chunk for the entire iteration. The number of pages
present in this final compressed chunk will be equal to the total number
of valid PFNs specified by the +ve chunks.

At the sender side, compressed pages are inserted into the output stream
in the same order as they would have been if compression logic was absent.

Until last iteration, the BODY is sent in Format A, to maintain live
migration compatibility with receivers of older Xen versions.
At the last iteration, if Remus compression was enabled, the sender sends
a trigger, XC_SAVE_ID_ENABLE_COMPRESSION to tell the receiver to parse the
BODY in Format B from the next iteration onwards.

An example sequence of chunks received in Format B:
    +16                              +ve chunk
    unsigned long[16]                PFN array
    +100                             +ve chunk
    unsigned long[100]               PFN array
    +50                              +ve chunk
    unsigned long[50]                PFN array

    XC_SAVE_ID_COMPRESSED_DATA       TAG
      N                              Length of compressed data
      N bytes of DATA                Decompresses to 166 pages

    XC_SAVE_ID_*                     other xc save chunks
    0                                END BODY TAG

Corner case with checkpoint compression:
    At sender side, after pausing the domain, dirty pages are usually
  copied out to a temporary buffer. After the domain is resumed,
  compression is done and the compressed chunk(s) are sent, followed by
  other XC_SAVE_ID_* chunks.
    If the temporary buffer gets full while scanning for dirty pages,
  the sender stops buffering of dirty pages, compresses the temporary
  buffer and sends the compressed data with XC_SAVE_ID_COMPRESSED_DATA.
  The sender then resumes the buffering of dirty pages and continues
  scanning for the dirty pages.
    For e.g., assume that the temporary buffer can hold 4096 pages and
  there are 5000 dirty pages. The following is the sequence of chunks
  that the receiver will see:

    +1024                       +ve chunk
    unsigned long[1024]         PFN array
    +1024                       +ve chunk
    unsigned long[1024]         PFN array
    +1024                       +ve chunk
    unsigned long[1024]         PFN array
    +1024                       +ve chunk
    unsigned long[1024]         PFN array

    XC_SAVE_ID_COMPRESSED_DATA  TAG
     N                          Length of compressed data
     N bytes of DATA            Decompresses to 4096 pages

    +4                          +ve chunk
    unsigned long[4]            PFN array

    XC_SAVE_ID_COMPRESSED_DATA  TAG
     M                          Length of compressed data
     M bytes of DATA            Decompresses to 4 pages

    XC_SAVE_ID_*                other xc save chunks
    0                           END BODY TAG

    In other words, XC_SAVE_ID_COMPRESSED_DATA can be interleaved with
  +ve chunks arbitrarily. But at the receiver end, the following condition
  always holds true until the end of BODY PHASE:
   num(PFN entries +ve chunks) >= num(pages received in compressed form)

TAIL PHASE
----------

Content differs for PV and HVM guests.

HVM TAIL:

 "Magic" pages:
    uint64_t         : I/O req PFN
    uint64_t         : Buffered I/O req PFN
    uint64_t         : Store PFN
 Xen HVM Context:
    uint32_t         : Length of context in bytes
    bytes            : Context data
 Qemu context:
    char[21]         : Signature:
      "QemuDeviceModelRecord" : Read Qemu save data until EOF
      "DeviceModelRecord0002" : uint32_t length field followed by that many
                                bytes of Qemu save data
      "RemusDeviceModelState" : Currently the same as "DeviceModelRecord0002".

PV TAIL:

 Unmapped PFN list   : list of all the PFNs that were not in map at the close
    unsigned int     : Number of unmapped pages
    unsigned long[]  : PFNs of unmapped pages

 VCPU context data   : A series of VCPU records, one per present VCPU
                       Maximum and present map supplied in XC_SAVE_ID_VCPUINFO
    bytes:           : VCPU context structure. Size is determined by size
                       provided in extended-info header
    bytes[128]       : Extended VCPU context (present IFF "extv" block
                       present in extended-info header)

 Shared Info Page    : 4096 bytes of shared info page
"""

CHUNK_end                       = 0
CHUNK_enable_verify_mode        = -1
CHUNK_vcpu_info                 = -2
CHUNK_hvm_ident_pt              = -3
CHUNK_hvm_vm86_tss              = -4
CHUNK_tmem                      = -5
CHUNK_tmem_extra                = -6
CHUNK_tsc_info                  = -7
CHUNK_hvm_console_pfn           = -8
CHUNK_last_checkpoint           = -9
CHUNK_hvm_acpi_ioports_location = -10
CHUNK_hvm_viridian              = -11
CHUNK_compressed_data           = -12
CHUNK_enable_compression        = -13
CHUNK_hvm_generation_id_addr    = -14
CHUNK_hvm_paging_ring_pfn       = -15
CHUNK_hvm_monitor_ring_pfn      = -16
CHUNK_hvm_sharing_ring_pfn      = -17
CHUNK_toolstack                 = -18
CHUNK_hvm_ioreq_server_pfn      = -19
CHUNK_hvm_nr_ioreq_server_pages = -20

chunk_type_to_str = {
    CHUNK_end                       : "end",
    CHUNK_enable_verify_mode        : "enable_verify_mode",
    CHUNK_vcpu_info                 : "vcpu_info",
    CHUNK_hvm_ident_pt              : "hvm_ident_pt",
    CHUNK_hvm_vm86_tss              : "hvm_vm86_tss",
    CHUNK_tmem                      : "tmem",
    CHUNK_tmem_extra                : "tmem_extra",
    CHUNK_tsc_info                  : "tsc_info",
    CHUNK_hvm_console_pfn           : "hvm_console_pfn",
    CHUNK_last_checkpoint           : "last_checkpoint",
    CHUNK_hvm_acpi_ioports_location : "hvm_acpi_ioports_location",
    CHUNK_hvm_viridian              : "hvm_viridian",
    CHUNK_compressed_data           : "compressed_data",
    CHUNK_enable_compression        : "enable_compression",
    CHUNK_hvm_generation_id_addr    : "hvm_generation_id_addr",
    CHUNK_hvm_paging_ring_pfn       : "hvm_paging_ring_pfn",
    CHUNK_hvm_monitor_ring_pfn      : "hvm_monitor_ring_pfn",
    CHUNK_hvm_sharing_ring_pfn      : "hvm_sharing_ring_pfn",
    CHUNK_toolstack                 : "toolstack",
    CHUNK_hvm_ioreq_server_pfn      : "hvm_ioreq_server_pfn",
    CHUNK_hvm_nr_ioreq_server_pages : "hvm_nr_ioreq_server_pages",
}

# Up to 1024 pages (4MB) at a time
MAX_BATCH = 1024

# Maximum #VCPUs currently supported for save/restore
MAX_VCPU_ID = 4095


"""
Libxl:

Legacy "toolstack" record layout:

Version 1:
  uint32_t version
  QEMU physmap data:
    uint32_t count
    libxl__physmap_info * count

The problem is that libxl__physmap_info was declared as:

struct libxl__physmap_info {
    uint64_t phys_offset;
    uint64_t start_addr;
    uint64_t size;
    uint32_t namelen;
    char name[];
};

Which has 4 bytes of padding at the end in a 64bit build, thus not the
same between 32 and 64bit builds.

Because of the pointer arithmatic used to construct the record, the 'name' was
shifted up to start at the padding, leaving the erronious 4 bytes at the end
of the name string, after the NUL terminator.

Instead, the information described here has been changed to fit in a new
EMULATOR_XENSTORE_DATA record made of NUL terminated strings.
"""



# -*- coding: utf-8 -*-

import os

def open_file_or_fd(val, *argl, **kwargs):
    """
    If 'val' looks like a decimal integer, open it as an fd.  If not, try to
    open it as a regular file.
    """

    fd = -1
    try:
        # Does it look like an integer?
        fd = int(val, 10)
    except ValueError:
        pass

    # Try to open it...
    if fd != -1:
        return os.fdopen(fd, *argl, **kwargs)
    else:
        return open(val, *argl, **kwargs)



#
#LiloConf.py
#

from __future__ import print_function, absolute_import

import sys, re, os
import logging
from . import GrubConf

class LiloImage(object):
    def __init__(self, lines, path):
        self.reset(lines, path)

    def __repr__(self):
        return ("title: %s\n"
                "  root: %s\n"
                "  kernel: %s\n"
                "  args: %s\n"
                "  initrd: %s\n" %(self.title, self.root, self.kernel,
                                   self.args, self.initrd))
    def reset(self, lines, path):
        self._initrd = self._kernel = self._readonly = None
        self._args = ""
        self.title = ""
        self.lines = []
        self.path = path
        self.root = ""
        for line in lines:
            self.set_from_line(line)

    def set_from_line(self, line, replace = None):
        (com, arg) = GrubConf.grub_exact_split(line, 2)

        if com in self.commands:
            if self.commands[com] is not None:
                setattr(self, self.commands[com], re.sub('^"(.+)"$', r"\1", arg.strip()))
            else:
                logging.info("Ignored image directive %s" %(com,))
        else:
            logging.warning("Unknown image directive %s" %(com,))

        # now put the line in the list of lines
        if replace is None:
            self.lines.append(line)
        else:
            self.lines.pop(replace)
            self.lines.insert(replace, line)

    def set_kernel(self, val):
        self._kernel = (None, self.path + "/" + val)
    def get_kernel(self):
        return self._kernel
    kernel = property(get_kernel, set_kernel)

    def set_initrd(self, val):
        self._initrd = (None, self.path + "/" + val)
    def get_initrd(self):
        return self._initrd
    initrd = property(get_initrd, set_initrd)

    def set_args(self, val):
        self._args = val
    def get_args(self):
        args = self._args
        if self.root:
            args += " root=" + self.root
        if self.readonly:
            args += " ro"
        return args
    args = property(get_args, set_args)

    def set_readonly(self, val):
        self._readonly = 1
    def get_readonly(self):
        return self._readonly
    readonly = property(get_readonly, set_readonly)

    # set up command handlers
    commands = { "label": "title",
                 "root": "root",
                 "rootnoverify": "root",
                 "image": "kernel",
                 "initrd": "initrd",
                 "append": "args",
                 "read-only": "readonly",
                 "chainloader": None,
                 "module": None}

class LiloConfigFile(object):
    def __init__(self, fn = None):
        self.filename = fn
        self.images = []
        self.timeout = -1
        self._default = 0

        if fn is not None:
            self.parse()

    def parse(self, buf = None):
        if buf is None:
            if self.filename is None:
                raise ValueError("No config file defined to parse!")

            f = open(self.filename, 'r')
            lines = f.readlines()
            f.close()
        else:
            lines = buf.split("\n")

        path = os.path.dirname(self.filename)
        img = []
        for l in lines:
            l = l.strip()
            # skip blank lines
            if len(l) == 0:
                continue
            # skip comments
            if l.startswith('#'):
                continue
            # new image
            if l.startswith("image"):
                if len(img) > 0:
                    self.add_image(LiloImage(img, path))
                img = [l]
                continue

            if len(img) > 0:
                img.append(l)
                continue

            (com, arg) = GrubConf.grub_exact_split(l, 2)
            if com in self.commands:
                if self.commands[com] is not None:
                    setattr(self, self.commands[com], arg.strip())
                else:
                    logging.info("Ignored directive %s" %(com,))
            else:
                logging.warning("Unknown directive %s" %(com,))

        if len(img) > 0:
            self.add_image(LiloImage(img, path))

    def hasPassword(self):
        return False

    def hasPasswordAccess(self):
        return True

    def add_image(self, image):
        self.images.append(image)

    def new_image(self, title, lines):
        # LiloImage constructor doesn't have title but since path
        # is being used by get_{kernel|initrd} functions we pass
        # empty string rather than None (see lines above)
        return LiloImage(lines, "")

    def _get_default(self):
        for i in range(len(self.images)):
            if self.images[i].title == self._default:
                return i
        return 0
    def _set_default(self, val):
        self._default = val
    default = property(_get_default, _set_default)

    commands = { "default": "default",
                 "timeout": "timeout",
                 "prompt": None,
                 "relocatable": None,
                 }

if __name__ == "__main__":
    if len(sys.argv) < 2:
        raise RuntimeError("Need a lilo.conf to read")
    g = LiloConfigFile(sys.argv[1])
    for i in g.images:
        print(i) #, i.title, i.root, i.kernel, i.args, i.initrd
    print(g.default)
